@@ -1,7 +1,9 @@
 #!/usr/bin/python3.8
 
 import argparse
+import base64
 import binascii
+import blake3
 import datetime
 import hashlib
 import dockerfile
@@ -16,12 +18,15 @@ import sseclient
 import subprocess
 import sys
 import tempfile
+import time
 
+from fastcdc import fastcdc
+from pathlib import Path
+from stat import ST_MODE
 from urllib.parse import urlparse
 
 VISOR_DIR = os.path.expanduser('~/.visor')
 VISOR_CONFIG = os.path.join(VISOR_DIR, 'config.json')
-VISOR_TRANSMIT_DIR = os.path.join(VISOR_DIR, 'transmit')
 
 # TODO: when running lots of builds in parallel, it seems like duplicate network
 # ids can be created, despite the locks on _create_namespaced_network()
@@ -126,7 +131,6 @@ def do_login(args):
   config['auths'][args.url]['sshPort'] = resp['sshPort']
   config['auths'][args.url]['sshUsername'] = resp['username']
   config['auths'][args.url]['sshKey'] = resp['key']
-  config['auths'][args.url]['sshTransmitDir'] = resp['repoPath']
   config['defaultServer'] = args.url
   os.makedirs(VISOR_DIR, exist_ok=True)
   with open(VISOR_CONFIG, 'w') as f:
@@ -220,6 +224,52 @@ def find_context_files(visorfile):
   files = [os.path.relpath(x, start=context_dir) for x in files]
   return files
 
+# TODO: Using small json names in order to minimize transport size. Use a
+# binary transport instead so the code remains readable
+# t - type
+# v - value
+# p - permissions
+def create_mapped_file(filemap, path):
+  parts = Path(path).parts
+  curpath = ''
+  for part in parts[:-1]:
+    curpath = os.path.join(curpath, part)
+    if part in filemap:
+      filemap = filemap[part]['v']
+      continue
+    filemap[part] = {
+      't': 'd',
+      'v': {},
+      'p': os.stat(curpath)[ST_MODE]
+    }
+    filemap = filemap[part]['v']
+  filemap[parts[-1]] = {}
+  return filemap[parts[-1]]
+
+def chunkify(files):
+  chunks = {}
+  filemap = {
+    't': 'd',
+    'v': {},
+  }
+  for f in files:
+    if os.path.islink(f):
+      mapped_file = create_mapped_file(filemap['v'], f)
+      mapped_file['t'] = 'l'
+      mapped_file['v'] = os.readlink(f)
+      mapped_file['p'] = os.stat(f)[ST_MODE]
+      continue
+    if os.path.isdir(f):
+      raise Exception(f'Transmitting directories not supported {f}')
+    mapped_file = create_mapped_file(filemap['v'], f)
+    mapped_file['t'] = 'f'
+    mapped_file['v'] = []
+    mapped_file['p'] = os.stat(f)[ST_MODE]
+    for chunk in fastcdc(f, 2**11, 2**14, 2**15, hf=blake3.blake3, fat=True):
+      mapped_file['v'].append(chunk.hash)
+      chunks[chunk.hash] = chunk.data
+  return filemap, chunks
+
 def transmit_context(visorfile, archive, server_url=None):
   server_url = verify_login(server_url)
   c = config['auths'][server_url]
@@ -227,31 +277,19 @@ def transmit_context(visorfile, archive, server_url=None):
     f.write(c['sshKey'])
     f.flush()
     with tempfile.NamedTemporaryFile('w', delete=False) as g:
-      g.write('\n'.join(find_context_files(visorfile)))
-      g.flush()
-      env = os.environ.copy()
-      env['RESTIC_PASSWORD'] = 'asdf'
-      ssh = ' '.join([
-        'ssh',
-        '-i', f.name,
-        f'{c["sshUsername"]}@{c["sshHost"]}',
-        '-p', str(c["sshPort"]),
-        '-s',
-        'sftp',
-      ])
-      args = ' '.join([
-        os.path.join(os.path.dirname(__file__), 'restic'),
-        'backup',
-        '-r', f'sftp::{c["sshTransmitDir"]}',
-        '--tag', archive,
-        '--files-from', g.name,
-        '--no-lock',
-        '-o', f'sftp.command="{ssh}"',
-      ])
-      print(args)
-      subprocess.check_call(
-        args, env=env, cwd=find_context_dir(visorfile), shell=True
+      files = find_context_files(visorfile)
+      filemap, chunks = chunkify(files)
+      missing_chunks = safe_non_stream_api_call(
+        'post', 'transmitmeta', json={'filemap': filemap}
+      )['missing']
+      safe_non_stream_api_call(
+        'post',
+        'transmit',
+        json={
+          'chunks': { x: base64.b64encode(chunks[x]) for x in missing_chunks }
+        }
       )
+  return filemap
 
 def do_build_base(args):
   safe_stream_api_call(
@@ -270,8 +308,7 @@ def do_build(args):
   visorfile = args.visorfile
   if os.path.isdir(visorfile):
     visorfile = os.path.join(visorfile, 'Visorfile')
-  transmit_context(visorfile, archive)
-  context_dir = find_context_dir(visorfile)
+  filemap = transmit_context(visorfile, archive)
   print('Done sending build context\n')
 
   buildenvs = {}
@@ -279,6 +316,7 @@ def do_build(args):
     a, b = x.split('=', 1)
     buildenvs[a] = b
 
+  context_dir = find_context_dir(visorfile)
   safe_stream_api_call(
     'post',
     'build',
@@ -287,6 +325,10 @@ def do_build(args):
       'visorfileRelpath': os.path.relpath(visorfile, start=context_dir),
       'keepRunning': args.keep_running,
       'buildenvs': buildenvs,
+      # TODO: transmitting the filemap once here and once above in the
+      # transmitmeta. Could keep something on the server so we only have to
+      # transmit once. The file metadata can get pretty big for lots of files.
+      'filemap': filemap,
     },
   )
 
